/* APACHETOP
**
**
*/
#include "apachetop.h"

struct gstat gstats;
time_t now;

Circle *c;
Timed_Circle *tc;
Hits_Circle *hc;

map *um, /* urlmap */
    *im, /* ipmap */
    *hm, /* hostmap */
    *rm; /* referrermap */

struct config cf;

WINDOW *win;

int dprintf(const char *fmt, ...)
{
	FILE *d;
        va_list args;

	if (cf.debug && (d = fopen(DEBUG_OUTPUT, "a")))
	{
        	va_start(args, fmt);
        	vfprintf(d, fmt, args);
		fclose(d);
        	va_end(args);
	}

        return 0;
}

int recordstats(struct logbits l)
{
	int t;

	/* global stats */
	if (gstats.alltime.first == 0) gstats.alltime.first = now;
	gstats.alltime.last = now;
	gstats.alltime.reqcount++;
	gstats.alltime.bytecount += l.bytes;

	/* for this return code, increment */
	t = (int)(l.retcode/100);
	gstats.r_codes[t].reqcount++;
	gstats.r_codes[t].bytecount += l.bytes;

	/* url and ip specific stuff was done in the parser */

	return 0;
}

static void finish(int s)
{
	endwin();

	exit(0);
}

int main(int argc, char *argv[])
{
	int fd, fd_in, buflen, ch, x;
	char buf[32768], *bufcp, *nl;
	int is_sp = 0;
	time_t last_display = 0;
	struct logbits lb;
	struct timespec stv;
	bool seen_h = false, seen_t = false;
	extern char *optarg;
#if HAVE_LIBSPREAD
	char private_group[MAX_GROUP_NAME];
#endif
#if HAVE_KQUEUE
	int kq, nev;
	struct kevent kev, ev;
#endif

#if HAVE_PTHREAD
	pthread_t key_thread; /* monitors keyboard */
#endif

	LogParser *p;
	CommonLogParser *cmmp; //AtopLogParser *atpp;

	stv.tv_sec = 0; stv.tv_nsec = 100000;

	/* some init stuff will need to know the time */
	now = time(NULL);

	/* set up initial configuration {{{ */
	cf.debug = true;
	cf.current_display_size = 0;
	cf.logfile = strdup(DEFAULT_LOGFILE);
	cf.logtype = DEFAULT_LOGTYPE;
	cf.circle_size = DEFAULT_CIRCLE_SIZE;
	cf.circle_mode = DEFAULT_CIRCLE_MODE;
	cf.sort = DEFAULT_SORT;
	cf.refresh_delay = DEFAULT_REFRESH_DELAY;
	cf.display_mode = DEFAULT_DISPLAY_MODE;
	cf.display_paused = false;
	cf.keep_querystring = false;
	cf.lowercase_urls = false;
	cf.keep_segments = 0;
	cf.preserve_ref_protocol = 0;
	cf.do_immed_display = false;
	/* }}} */

	/* process commandline */
	while ((ch = getopt(argc, argv, "f:t:H:T:hqls:pr:")) != -1)
	{
		switch(ch)
		{
			case 'f':
				free(cf.logfile);
				cf.logfile = strdup(optarg);
				break;
			case 't':
				if (strcmp(optarg, "common") == 0)
					cf.logtype = LOG_COMMON;
				else if (strcmp(optarg, "combined") == 0)
					cf.logtype = LOG_COMBINED;
				else if (strcmp(optarg, "atop") == 0)
					cf.logtype = LOG_ATOP;
				break;
			case 'T':
				x = atoi(optarg);
				if (x > 0)
				{
					cf.circle_mode = TIMED_CIRCLE;
					cf.circle_size = x;
					seen_t = true;
				}
				break;
			case 'H':
				x = atoi(optarg);
				if (x > 0)
				{
					cf.circle_mode = HITS_CIRCLE;
					cf.circle_size = x;
					seen_h = true;
				}
				break;

			case 'q':
				cf.keep_querystring = true;
				break;

			case 'l':
				cf.lowercase_urls = true;
				break;

			case 's':
				x = atoi(optarg);
				if (x > 0)
					cf.keep_segments = x;
				break;

			case 'p':
				cf.preserve_ref_protocol = 1;
				break;

			case 'r':
				x = atoi(optarg);
				if (x > 0)
					cf.refresh_delay = x;
				break;

			case 'h': 
			case '?': 
				usage(argv[0]);
				exit(1);
				break;
		}
	}

	if (seen_t && seen_h)
	{
		fprintf(stderr, "%s: -t and -h are mutually exclusive."
		    "Specify only one.\n", argv[0]);
		exit(1);
	}
#if HAVE_LIBSPREAD	
	if(strchr(cf.logfile, '@') && strchr(cf.logfile, ':')) {
		char user[MAX_PRIVATE_NAME];
		char *cp = strchr(cf.logfile, ':');
		*(cp++) = (char) NULL;
		snprintf(user, sizeof(user), "atop-%05d", getpid());
		if (SP_connect(cf.logfile, user, 0, 0, &fd_in,
		               private_group) != ACCEPT_SESSION)
		{
			fprintf(stderr, "%s: could not connect to spread\n",
			    argv[0]);
			exit(1);
		}
		if (SP_join(fd_in, cp) != 0)
		{
			fprintf(stderr, "%s: could not join logging group\n",
			    argv[0]);
			exit(1);
		}
		is_sp = 1;
	}
	else
#endif
	if ((fd_in = open(cf.logfile, O_RDONLY)) == -1)
	{
		usage(argv[0]);
		fprintf(stderr, "%s: %s\n",
		    cf.logfile, strerror(errno));
		exit(1);
	}
	lseek(fd_in, 0, SEEK_END);

#if HAVE_KQUEUE
	if ((kq = kqueue()) < 0)
	{
		fprintf(stderr, "%s: cannot create kqueue: %s\n",
		    argv[0], strerror(errno));
		exit(1);
	}
	EV_SET(&kev, fd_in, EVFILT_READ, EV_ADD | EV_ENABLE, 0, 0, 0);
	if (kevent(kq, &kev, 1, NULL, 0, &stv) < 0)
	{
		fprintf(stderr, "%s: cannot create kevent: %s\n",
		    argv[0], strerror(errno));
		exit(1);
	}
#endif

	/* remember cf.circle_size hits */
	switch(cf.circle_mode)
	{
		default:
		case HITS_CIRCLE:
			hc = new Hits_Circle;
			hc->create(cf.circle_size);
			c = hc;
			break;
	
		case TIMED_CIRCLE:
			tc = new Timed_Circle;
			tc->create(cf.circle_size);
			c = tc;
			break;
	}

	switch(cf.logtype)
	{
		default:
		case LOG_COMMON:
		case LOG_COMBINED: /* CommonLogParser handles combined too */
			cmmp = new CommonLogParser;
			p = cmmp;
			break;
#if 0
/* this isn't written yet */
		case LOG_ATOP:
			atpp = new AtopLogParser;
			p = atpp;
			break;
#endif
	}

	/* maps auto-resize, just use a sane starting point */
	/* url string -> url hash map */
	um = new map; /* hashing happens internally */
	um->create(cf.circle_size);

	/* referrer string -> hash map */
	rm = new map;
	rm->create(cf.circle_size);

	/* ip string -> ip hash map */
	im = new map; 
	im->create(cf.circle_size);

	/* host string -> host hash map */
	hm = new map; 
	hm->create(cf.circle_size);

	//queue = new Queue;

	memset(&gstats, (char) NULL, sizeof(gstats));
	gstats.start = time(NULL);

	signal(SIGINT, finish);
	signal(SIGKILL, finish);

	initscr();
	win = newwin(0,0,0,0);
	nonl(); /* no NL->CR/NL on output */
	cbreak(); /* enable reading chars one at a time */
	noecho();
	keypad(win, true);

#if HAVE_PTHREAD
	nodelay(win, false);
	pthread_create(&key_thread, NULL, &key_processor, NULL);
#else
	nodelay(win, true);
#endif

	for(;;)
	{
		now = time(NULL);

#if !HAVE_PTHREAD
		if ((ch = wgetch(win)) != ERR)
			readKey(ch);
#endif

		/* only update display .. */
		if (cf.do_immed_display || /* if we've been asked to, or */
		    (!cf.display_paused && /* we're not paused and .. */
		     /* the refresh_delay has passed */
		     now - last_display >= cf.refresh_delay))
		{
			display();
			cf.do_immed_display = false;
			last_display = now; // remember our last update
		}

#if HAVE_KQUEUE
		x = kevent(kq, NULL, 0, &ev, 1, NULL);
		if (x == 0) continue; /* timeout */
		if (x < 0) break; /* error */

		/* we can determine which fd returned for us,
		 * and even how much data there is to read */
		
		/* when we're ready to monitor more than one file, we can
		 * make ev an array of structs, then cycle through each one
		*/
		fd = ev.ident;
		//read_amt = ev.data; /* we don't use this */

#else /* fall back */

		/* just sleep for a while */
		nanosleep(&stv, NULL);

		/* when we have more than one file..
		 * we'll have to check them all
		*/
		fd = fd_in;
#endif

		/* read the data */
		if(!is_sp)
		{
			buflen = read(fd, buf, sizeof(buf));

			if (buflen == 0)
				continue; /* no data */

			if (buflen < 0)
				break; /* error */
		}

#if HAVE_LIBSPREAD
		else
		{
			int16 mt = 0;
			char sender[MAX_GROUP_NAME];
			char tgroups[10][MAX_GROUP_NAME];
			int num_groups, st = 0, endian = 0;
			buflen = SP_receive(fd_in, &st, sender, 10,
			                       &num_groups, tgroups, &mt,
					       &endian, sizeof(buf)-1, buf);
			if (buflen <= 0)
			{
				fprintf(stderr, "error reading from spread\n");
				exit(1);
			}
			buf[buflen] = (char) NULL;
			//dprintf("%s", buf);
		}
#endif

		/* pass each line to logsplit */
		nl = bufcp = buf;
		while ((nl - buf) < buflen && nl)
		{
			bufcp = nl;
			/* find the end of this line */
			if ((nl = strchr(bufcp, '\n')))
			{
				*nl = (char) NULL;
				++nl;

				if (p->parse(bufcp, &lb) == 0)
				{
					/* insert into circle */
					c->insert(lb);

					/* record stats */
					recordstats(lb);
				}
			}

			// next line..
		}
	}

	/* NOTREACHED */
	return 0;
}
#if 0
/* resolver stuff, unused {{{ */
/* background thread that watches ipmap and resolves IPs in it to hosts if
 * they don't have an entriy in the hostmap; and vice versa.
 *
 * entries between the maps are linked via their pos; they have the same one.
*/
void *resolver(void *)
{
	int rc;
	struct qresult *r;
	struct timeval stv;

	/* some tmp buffers for resolve() to fill in */
	char *host, ip[256];
	int ipl;

	stv.tv_sec = 2;
	stv.tv_usec = 0;

	/* run sequence:
	**
	** every few seconds:
	** loop through ipmap (write walk()) getting the pos for each one
	** reverse that pos in hostmap and see if it's NULL
	**  if it is, do gethost crap to resolve it
	**  store result
	*/
	do
	{
		/* wait for entries to appear in the queue, then process */
		if (queue->entries() && (rc = queue->pop( (void **)&r )) == 0)
		{
			dprintf("got %s to check\n", r->src);

			resolve(r->src, &ipl, (char **)&ip, &host);

			dprintf("host of that is %s\n", host);
			/* do what now? */

			free(r->src);
			free(r);
		}
		else
		{
			dprintf("sleep\n");
			/* zzz */
			select(0, NULL, NULL, NULL, &stv);
		}
	}
	while (queue->entries() > 0 || main_done == 0);

	return 0;
}
/* given an unknown source, populate the ipl, ip, and host fields, where
 * possible; 
 * host must be a pointer, ip must be a char[].
*/
int resolve(char *src, int *ipl, char **ip, char **host)
{
	in_addr_t f;
	struct hostent *he;

	if ((f = inet_addr(src)) == INADDR_NONE)
	{
		/* its a host */
		if ((he = gethostbyname(src)))
		{
			*host = src;
			inet_ntop(AF_INET, he->h_addr, *ip, sizeof(*ip));
			*ipl = inet_addr(*ip);
		}
		else
		{
			/* it didn't resolve; this shouldn't happen because
			 * Apache shouldn't put duff DNS in its logs */
			*host = src;
			*ipl = 0;
			*ip = NULL;
		}
	}
	else
	{
		/* its an IP */
		*ipl = f;
		if ((he = gethostbyaddr((char *)&f, sizeof(f), AF_INET)))
		{
			*host = he->h_name;
			inet_ntop(AF_INET, he->h_addr, *ip, sizeof(*ip));
		}
		else
		{
			/* no reverse, this is fine */
			*host = NULL;
			*ip = src;
		}
	}
	//printf("host: %s ip: %s ipl: %d\n", *host, *ip, *ipl);

	return 0;
}
/* }}} */
#endif

#if HAVE_PTHREAD
void *key_processor(void *)
{
	/* sit in getch() and pass to readKey when we get something */
	int ch;

	for( ;; )
	{
		ch = wgetch(win);
		readKey(ch);
	}
}
#endif

int readKey(int ch)
{
	/* got a keypress, best process it */
	dprintf("got a %d\n", ch);
	switch(ch)
	{
		case 'p': /* (un)pause display */
			cf.display_paused = !cf.display_paused;
			if (cf.display_paused)
			{
				/* tell the user we're paused */
				attron(A_REVERSE);
				mvprintw(0, 60, "paused");
				attroff(A_REVERSE);
				refresh();
			}
			else
			{
				/* turning off pause forces an update */
				move(0, 60); /* clear "paused" */
				clrtoeol(); /* time will get redrawn */
				display();
			}
			break;
	
		case 'r': /* sorting */
			cf.sort = SORT_REQCOUNT;
			cf.do_immed_display = true;
			break;
		case 'R':
			cf.sort = SORT_REQPERSEC;
			cf.do_immed_display = true;
			break;
		case 'b':
			cf.sort = SORT_BYTECOUNT;
			cf.do_immed_display = true;
			break;
		case 'B':
			cf.sort = SORT_BYTESPERSEC;
			cf.do_immed_display = true;
			break;
#if 0
		case 's': /* new refresh delay */ 
#ifndef SOLARIS /* linking against readline is failing on Solaris atm */
			endwin();
			char *t;
			int nd;
			t = readline("Seconds to Delay: ");
			if ((nd = atoi(t)) != 0)
				cf.refresh_delay = nd;

			refresh();
#endif
			break;
#endif
		case 'd': /* display mode; urls or hosts */
			switch(cf.display_mode)
			{
			case DISPLAY_URLS:
				cf.display_mode = DISPLAY_IPS;
				break;
			case DISPLAY_IPS:
				cf.display_mode = DISPLAY_REFS;
				break;
			case DISPLAY_REFS:
				cf.display_mode = DISPLAY_URLS;
				break;
			}
			cf.do_immed_display = true;
			break;

		case KEY_UP:
			/* sanity checking for this is done in drawMarker() */
			cf.selected_item--;
			drawMarker();
			break;

		case KEY_DOWN:
			/* sanity checking for this is done in drawMarker() */
			cf.selected_item++;
			drawMarker();
			break;
		
		case KEY_REFRESH:
			refresh();
			break;

		case KEY_BREAK:
		case 'q':
			finish(0);
	}
	
	return 0;
}

void usage(char *cmd)
{
	fprintf(stderr, 
	    "Usage:\n"
	    "File options:\n"
	    "  -f logfile  open logfile [%s]\n"
	    "  -t logtype  specify logfile type [common]\n"
	    "\n"
	    "URL/referrer munging options:\n"
	    "  -q          keep query strings [no]\n"
	    "  -l          lowercase all URLs [no]\n"
	    "  -s num      keep num path segments of URL [all]\n"
	    "  -p          preserve protocol at front of referrers [no]\n"
	    "\n"
	    "Stats options:\n"
	    "  Supply up to one of the following two. default: [-%c %d]\n"
	    "  -H hits     remember stats for this many hits\n"
	    "  -T secs     remember stats for this many seconds\n"
	    "\n"
	    "  -r secs     refresh delay in seconds [%d]\n"
	    "\n"
	    "  -h          this help\n"
	    "\n"
	    ,
	    DEFAULT_LOGFILE,
	    DEFAULT_CIRCLE_MODE, DEFAULT_CIRCLE_SIZE,
	    DEFAULT_REFRESH_DELAY
	    );

	return;
}
