#ifndef _APACHETOP_H_
#define _APACHETOP_H_

#if HAVE_CONFIG_H
# include "../config.h"
#endif

#include <stdio.h>
#include <errno.h>
#include <ctype.h>
#include <string.h>

#if HAVE_FCNTL_H
# include <fcntl.h>
#else
# define O_RDONLY 0x0000
  extern "C" int open(const char *, int, ...);
#endif

#include <signal.h>

#if HAVE_STDLIB_H
# include <stdlib.h>
#else
extern "C"
{
  void exit(int);
  void abort(void);
  void free(void *);
  int atoi(const char *);
  void *malloc(size_t), *realloc(void *, size_t), *calloc(size_t, size_t);
}
#endif

#if HAVE_STDARG_H
#include <stdarg.h>
#endif

#if HAVE_UNISTD_H
# include <unistd.h>
#else
extern "C"
{
  extern char *optarg;
  extern int optreset;
  int getopt(int, char * const [], const char *);

  ssize_t read(int, void *, size_t);
}
#endif

#if HAVE_SYS_TYPES_H
# include <sys/types.h>
#endif

#if TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# if HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif

#if HAVE_SYS_SOCKET_H
# include <sys/socket.h>
#endif

#if HAVE_NETINET_IN_H
# include <netinet/in.h>
#endif

#if HAVE_ARPA_INET_H
# include <arpa/inet.h>
#else
  extern "C" in_addr_t inet_addr (const char *);
#endif

#if HAVE_NETDB_H
# include <netdb.h>
#endif

#include <curses.h>

#if HAVE_LIBSPREAD
# include <sp.h>
#endif

#if HAVE_KQUEUE
# include <sys/event.h>
#endif

#if HAVE_PTHREAD
# include <pthread.h>
#endif


#define getMIN(a,b) (a < b ? a : b)
#define getMAX(a,b) (a > b ? a : b)

struct config {
#define SORT_REQCOUNT 1
#define SORT_REQPERSEC 2
#define SORT_BYTECOUNT 3
#define SORT_BYTESPERSEC 4
	short sort;
	short refresh_delay;

	short selected_item; /* screen position our marker is at */
	unsigned int selected_item_pos; /* which url/ip/refpos it relates to */
	short selected_item_mode; /* is it url/ip/ref pos? */

	short current_display_size; /* how many lines we're displaying */

	char *logfile;
	short logtype; /* types defined in log.h */

#define TIMED_CIRCLE 'T'
#define HITS_CIRCLE 'H'
	int circle_mode;

	int circle_size;

#define DISPLAY_URLS 1
#define DISPLAY_IPS 2
#define DISPLAY_REFS 3
	short display_mode;

	bool display_paused;
	bool do_immed_display; /* signals a screen update is req'd */

	/* if nonzero, show referrer stats for the given url_pos */
	int display_referrers;

	/* filter */
	char *filter;
	short filter_mode; /* uses DISPLAY_ defines */

	/* url munging */
	bool keep_querystring;
	bool lowercase_urls;
	unsigned short keep_segments;
	bool preserve_ref_protocol;

	bool debug;
};

/* parse*() and resolver() use this to push and pull items off the Queue */
struct qresult
{
	char *src;
	int pos;
};

struct gstat {
	time_t start; /* when did we start */

	/* space for 1xx-5xx return codes */
	struct r_code {
		double bytecount;
		double reqcount;
	} r_codes[6];

	/* space for counting global hits and bytes per sec */
	struct global {
		double bytecount;
		double reqcount;
		time_t first, last;
	} alltime;

};

#include "ohtbl.h"
#include "map.h"
#include "hashes.h"
#include "circle.h"
#include "hits_circle.h"
#include "timed_circle.h"
#include "display.h"
#include "log.h"


#define JAN 281
#define FEB 269
#define MAR 288
#define APR 291
#define MAY 295
#define JUN 301
#define JUL 299
#define AUG 285
#define SEP 296
#define OCT 294
#define NOV 307
#define DEC 268

#define DEBUG_OUTPUT "/tmp/atop.debug"

#ifndef DEFAULT_LOGFILE
# define DEFAULT_LOGFILE "/var/httpd/apache_log"
#endif
#define DEFAULT_LOGTYPE LOG_COMMON
#define DEFAULT_CIRCLE_SIZE 30
#define DEFAULT_CIRCLE_MODE TIMED_CIRCLE
#define DEFAULT_SORT SORT_REQCOUNT
#define DEFAULT_REFRESH_DELAY 5
#define DEFAULT_DISPLAY_MODE DISPLAY_URLS

/* if the layout of the display changes, these need updating */
#define COLS_RESERVED 25
#define LINES_RESERVED 7

int dprintf(const char *fmt, ...);

void *resolver(void *);
int resolve(char *src, int *ipl, char **ip, char **host);

#if HAVE_PTHREAD
void *key_processor(void *);
#endif
int readKey(int ch);
void usage(char *cmd);

#endif
