#include "apachetop.h"

extern struct gstat gstats;
extern time_t now;

extern Circle *c;

extern map *um, /* urlmap */
           *im, /* ipmap */
           *hm, /* hostmap */
           *rm; /* referrermap */

extern struct config cf;

float readableNum(double num, char *suffix)
{
	#define MAX_IN_KB (1024*1024)*10

	if (num < MAX_IN_KB)
	{
		*suffix = 'K';
		return (float)num/1024;
	}

	*suffix = 'M';
	return (float)num/(1024*1024);
}

void display()
{
	unsigned int diff, d = 0, h = 0, m = 0, s = 0;
	float bytes, ftmp;
	int itmp, x, disp;
	char ctmp[80], bytes_suffix, *cptmp;
	struct tm *tm;
	struct config scf; /* for copying cf, see comment below at memcpy */
	
	/* for easy reference during walk() */
	map *map; unsigned int pos;

	/* walk() pointer */
	struct logbits *lb;

	/* last hit and uptime */
	tm = localtime(&gstats.alltime.last);
	strftime(ctmp, sizeof(ctmp), "%H:%M:%S", tm);
	diff = (unsigned int)difftime(now, gstats.start);
	if (diff > 86399) diff -= ((d = diff / 86400)*86400);
	if (diff > 3599) diff -= ((h = diff / 3600)*3600);
	if (diff > 59) diff -= ((m = diff / 60)*60);
	s = diff;
	mvprintw(0, 0,
	  "last hit: %s        atop runtime: %2d days, %02d:%02d:%02d",
	  ctmp, d, h, m, s );

	/* current time */
	tm = localtime(&now);
	strftime(ctmp, sizeof(ctmp), "%H:%M:%S", tm);
	mvprintw(0, 71, "%s", ctmp);

	//All: 1,140,532 requests (39.45/sec),  999,540,593 bytes (857,235/sec) 
	bytes = readableNum(gstats.alltime.bytecount, &bytes_suffix);
	ftmp = getMAX(now-gstats.alltime.first, 1); /* divide-by-zero hack */
	attron(A_BOLD);
	mvprintw(1, 0, 
	  "All: %12.0f reqs (%6.1f/sec)  %11.1f%c (%6.1fK/sec)  (%7.1fK/req)",
	  gstats.alltime.reqcount,
	  gstats.alltime.reqcount/ftmp,
	  bytes, bytes_suffix,
	  (gstats.alltime.bytecount/1024)/ftmp,
	  (float)(gstats.alltime.bytecount/1024)/
	      getMAX(gstats.alltime.reqcount, 1));
	attroff(A_BOLD);


	// 2xx 1,604,104 (95%) 3xx 1,000,000 ( 3%) 4xx 1,000,000 ( 1%) 
	// 5xx 1,000,000 ( 1%)
	ftmp = gstats.r_codes[2].reqcount + gstats.r_codes[3].reqcount+
	       gstats.r_codes[4].reqcount + gstats.r_codes[5].reqcount;
	if (ftmp == 0) ftmp = 1; /* avoid NaN with no hits */
	mvprintw(2, 0,
	  "2xx: %7.0f (%4.1f%%) 3xx: %7.0f (%4.1f%%) "
	  "4xx: %5.0f (%4.1f%%) 5xx: %5.0f (%4.1f%%) ",
	  gstats.r_codes[2].reqcount, (gstats.r_codes[2].reqcount/ftmp)*100,
	  gstats.r_codes[3].reqcount, (gstats.r_codes[3].reqcount/ftmp)*100,
	  gstats.r_codes[4].reqcount, (gstats.r_codes[4].reqcount/ftmp)*100,
	  gstats.r_codes[5].reqcount, (gstats.r_codes[5].reqcount/ftmp)*100);

	/* housecleaning on the circle, if its required in this class */
	c->updatestats();
	/* fetch the time of the first "recent" request */
	itmp = now - c->oldest();
	itmp = getMAX(itmp, 1); /* divide-by-zero hack */
	bytes = readableNum(c->getbytecount(), &bytes_suffix);
	attron(A_BOLD);
	mvprintw(3, 0,
	  "R (%3ds): %7.0f reqs (%6.1f/sec)  %11.1f%c (%6.1fK/sec)  (%7.1fK/req)",
	  itmp, c->getreqcount(),
	  ((float)c->getreqcount()/itmp),
	  bytes, bytes_suffix,
	  ((float)(c->getbytecount()/1024)/itmp),
	  (float)(c->getbytecount()/1024)/getMAX(c->getreqcount(), 1)
	  );
	attroff(A_BOLD);

	ftmp = c->getsummary(2) + c->getsummary(3) +
	       c->getsummary(4) + c->getsummary(5);
	if (ftmp == 0) ftmp = 1; /* avoid NaN with no hits */
	mvprintw(4, 0,
	  "2xx: %7.0f (%4.1f%%) 3xx: %7.0f (%4.1f%%) "
	  "4xx: %5.0f (%4.1f%%) 5xx: %5.0f (%4.1f%%) ",
	  c->getsummary(2), (c->getsummary(2)/ftmp)*100,
	  c->getsummary(3), (c->getsummary(3)/ftmp)*100,
	  c->getsummary(4), (c->getsummary(4)/ftmp)*100,
	  c->getsummary(5), (c->getsummary(5)/ftmp)*100);

	/* top $X urls or ips/hosts */

	/* make an array containing all we have, then sort it */

	OAHash item_hash;
	int item_used = 0;
	double items_size;
	struct peritem_stat *items = NULL, *item_ptr = NULL;

	items_size = c->getreqcount();

	item_hash.create(items_size*5);

	/* this is our array; the cast is because items_size is a double but
	 * I'm fairly sure, realistically, it'll never get high enough to be
	 * a problem, so uInt should be ok */
	items = (struct peritem_stat *)
	    calloc((unsigned int)items_size * sizeof(peritem_stat), 1);
	
	/* another thread may change the contents of cf while we're running,
	 * and it would be undesirable to have most of this change on us, so
	 * we make safe copies to use */
	memcpy(&scf, &cf, sizeof(struct config));
	
	/* pick a map to use */
	switch(scf.display_mode)
	{
		default:
		case DISPLAY_URLS: map = um; break;
		case DISPLAY_IPS: map = im; break;
		case DISPLAY_REFS: map = rm; break;
	}

	while(c->walk(&lb) != -1)
	{
		/* skip unused */
		if (lb == NULL)
			continue;
	
		/* FILTERS */

		/* set up some pointers, depending on cf.display_mode, so
		 * that we can refer to the url_pos or ip_pos via just pos,
		 * and the urlmap or ipmap via just map
		*/
		switch(scf.display_mode)
		{
			default:
			case DISPLAY_URLS: pos = lb->url_pos; break;
			case DISPLAY_IPS: pos = lb->ip_pos; break;
			case DISPLAY_REFS: pos = lb->ref_pos; break;
		}

		cptmp = map->reverse(pos);
		item_ptr = (struct peritem_stat *)item_hash.lookup(cptmp);
		
		if (item_ptr == NULL)
		{
			/* not seen it, make a new slot */
			item_ptr = &items[item_used];

			item_hash.insert(map->reverse(pos), &items[item_used]);
			item_ptr->item = pos;

			++item_used;
			item_ptr->last = now;
			item_ptr->first = lb->time;
		}

		/* update stats for this item in the array */
		++(item_ptr->reqcount);
		item_ptr->bytecount += lb->bytes;

		if (lb->time < item_ptr->first)
			item_ptr->first = lb->time;
	}

	/* we're done with this */
	item_hash.destroy();

	/* calculate some stats; timespan, kbps, rps */
	for(x = 0 ; x < item_used ; ++x)
	{
		item_ptr = &items[x];

		item_ptr->timespan = item_ptr->last - item_ptr->first;
		if (item_ptr->timespan == 0)
			item_ptr->timespan = 1;

		item_ptr->kbps = ((float)
		    (item_ptr->bytecount/1024)/item_ptr->timespan);
		item_ptr->rps = ((float)
		    item_ptr->reqcount/item_ptr->timespan);
	}

#if 0
	/* what are we showing in the table? we have a few options;
	 *
	 * top $X URLS or IPs/Hosts or Referrers
	 * detailed referrer stats for a given URL
	*/
	if (cf.display_mode == DISPLAY_URLS && cf.selected_item_pos)
	{
		/* detailed referrer stats for a given URL */
		clrtobot();
		mvprintw(LINES_RESERVED-1, 0, "REQS REQ/S    KB  KB/S URL");

		/* find the url_pos given by display_referrer in items */
		for (x = 0 ; x < item_used ; ++x)
		{
			item_ptr = &items[x];
			if (item_ptr->item == cf.selected_item_pos)
			{
				/* show stats for this line only */
				showDetail(item_ptr, 0);
				break;
			}
		}

		/* now display breakdown of referrers for that line */
		//displayReferrers();
		return;
	}
#endif

	/* top $X items; sort the array for display */
	if (item_used) shellsort(items, item_used, scf.sort);

	clrtobot();

	/* display something */

	/* a suitable header, at LINES_RESERVED-1
	 * (-1 because curses starts at zero, but I've started at 1) */
	switch(scf.display_mode)
	{
		case DISPLAY_URLS:
		mvprintw(LINES_RESERVED-1, 0, "REQS REQ/S    KB  KB/S URL");
		break;
	
		case DISPLAY_IPS:
		mvprintw(LINES_RESERVED-1, 0, "REQS REQ/S    KB  KB/S HOST");
		break;

		case DISPLAY_REFS:
		mvprintw(LINES_RESERVED-1, 0, "REQS REQ/S    KB  KB/S REFERRER");
		break;
	}

	disp = 0; /* count how many we've shown */
	for(x = 0, item_ptr = &items[0] ;
	    x < items_size  &&  disp < LINES-LINES_RESERVED ;
	    ++item_ptr, ++x)
	{
		/* translate a screen position into an url/ip/ref_pos,
		 * depending on current display mode. We can then use
		 * selected_item_pos to get info for the
		 * selected item if the user hits return to get it */
		if (cf.selected_item == x)
		{
			cf.selected_item_pos = item_ptr->item;
			cf.selected_item_mode = scf.display_mode;
		}
	
		/* skip empty tablespaces, even though none should exist */
		if (item_ptr->reqcount == 0)
			continue;

		/* render the line itself at position disp. map should be
		 * already set from earlier in this function */
		showDetail(item_ptr, disp, map);

		++disp;
	}
	free(items);

	cf.current_display_size = disp;

	/* if there were items, draw a marker */
	if (cf.current_display_size)
		drawMarker();
	else
	{
		/* come to rest just above the column headers */
		move(LINES_RESERVED-2,0);
		refresh();
	}
}

/* update position of asterisk next to URLs */
void drawMarker(void)
{
	if (cf.current_display_size == 0)
		return;

	/* ensure our marker isn't beyond the end of the list */
	if (cf.selected_item > cf.current_display_size-1)
		cf.selected_item = cf.current_display_size-1;
	
	/* or above the start of it */
	if (cf.selected_item < 0) cf.selected_item = 0;

	/* draw an asterisk next to the selected item and clear adjacent lines*/
	mvaddch(LINES_RESERVED+cf.selected_item-1, COLS_RESERVED-3, ' ');
	mvaddch(LINES_RESERVED+cf.selected_item, COLS_RESERVED-3, '*' | A_BOLD);
	mvaddch(LINES_RESERVED+cf.selected_item+1, COLS_RESERVED-3, ' ');

	/* come to rest just above the column headers */
	move(LINES_RESERVED-2,0);
	
	refresh();
}

/* show all referrers for the given item_ptr's urlpos */
void displayReferrers(struct peritem_stat *item_ptr)
{
}

/* render contents of item_ptr into a statistics line onscreen,
 * at offset vert_location; pull the text portion out of m */
void showDetail(struct peritem_stat *item_ptr, int vert_location, map *m)
{
	/* start drawing at LINES_RESERVED */
	mvprintw(LINES_RESERVED+vert_location, 0,
	    "%4.0f %5.*f %5.*f %5.*f %.*s",
	    item_ptr->reqcount,

	    ((item_ptr->rps > 99) ? 1 : 2),
	    item_ptr->rps,

	    /* scale KB display; if >1000K lose the decimal point */
	    ((item_ptr->bytecount > (999*1024)) ? 0 : 1),
	    ((float)item_ptr->bytecount/1024),
	    
	    /* scale KB/s display; if >1000K/s lose the decimal point */
	    ((item_ptr->kbps > 999) ? 0 : 1),
	    item_ptr->kbps,

	    (COLS-COLS_RESERVED /* width */), m->reverse(item_ptr->item));
}

void shellsort(struct peritem_stat *items, unsigned int size, int sorttype)
{
	bool done;
	unsigned int i, j, jmp = size;
	double i_c, j_c;

	struct peritem_stat tmp;

	while (jmp > 1)
	{
		jmp >>= 1;
		do
		{
			done = true;
			for (j = 0; j < (size-jmp); j++)
			{
				i = j + jmp;
				switch(sorttype)
				{
					case SORT_REQCOUNT:
						i_c = items[i].reqcount;
						j_c = items[j].reqcount;
						break;
					case SORT_REQPERSEC:
						i_c = items[i].rps;
						j_c = items[j].rps;
						break;
					case SORT_BYTECOUNT:
						i_c = items[i].bytecount;
						j_c = items[j].bytecount;
						break;
					case SORT_BYTESPERSEC:
						i_c = items[i].kbps;
						j_c = items[j].kbps;
						break;
				}
				
				if (i_c > j_c)
				{
#define P_S_S sizeof(struct peritem_stat)
					memcpy(&tmp, &items[i], P_S_S);
					memcpy(&items[i], &items[j], P_S_S);
					memcpy(&items[j], &tmp, P_S_S);

					done = false;
				}
			}
		} while (!done);
	}
}
